package gc.globalGC;

/**
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: WWC</p>
 * @author WeiJen Wang
 * @version 1.0
 */
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Hashtable;
import java.util.HashSet;
import java.util.Vector;
import java.util.Iterator;
import salsa.language.UniversalActor;
import salsa.language.ServiceFactory;
import salsa.naming.*;
import gc.*;

public class GBackTracer implements java.io.Serializable{
  private String myName;

  //The internal data structure for mark and sweep:
  // Let host be the host address.
  // Querying tracingList with 'host' gets a HashSet that contains
  // all the next traget actor references in this host
  private Hashtable tracingList;
  private int totalTargets=0;
  transient private LocalCollector gc;

  public GBackTracer(ActorMemory startPoint) {
    myName=startPoint.GCGetForwardList().getSelfRef();
    tracingList=new Hashtable();
    putData(startPoint.GCGetForwardList().getSelfRef());
    gc=ServiceFactory.getGC();
  }

  public static final int MIGRATION=0;
  public static final int SUCCESS=1;
  public static final int FAILURE=2;
  public static final int COMPETEION=3;
  public static final int ERROR=4;

  //This function returns the decision.
  // 0: no more local data, the actor will migrate to the next host
  // 1: no more data, the search is finished. All marked actors are garbage
  // 2: Cannot find relative data which means the tracing may not be garbage
  // 3: There is a competetion
  // -1: errors.
  public int oneHostBFSMark() {
    //lock the local garbage collector
    gc.occupy(myName);

    //get the local targets
    HashSet data=(HashSet)tracingList.get(getLocalLocation());
    Hashtable nonlocalSet=gc.getNonlocalReachableSet(myName);
    if (data!=null) {
      Iterator it=data.iterator();
      String target;

      //produce next targets to mark and sweep
      while (it.hasNext()) {
        target=(String)it.next();
        it.remove();
        if (nonlocalSet!=null) {
          ActorMemory targetMem=(ActorMemory)nonlocalSet.get(target);
          //if we find the target actor memory, mark it and get its inverse acquaintances
          //otherwise the suspect is not garbage
          if (targetMem!=null) {

            //if the actor has the global mark the actor marks, skip it,
            //otherwise if the mark is empty, keep backtracing
            // If there is another mark, there is a competetion.
            if (targetMem.isGlobalMarked().equals(myName)) {continue;}
            else if (targetMem.isGlobalMarked()==null){
              targetMem.colorGlobalMark(myName);
              for (Enumeration e=targetMem.GCGetinverseList().getRealInverseList().keys();e.hasMoreElements();) {
                String nextTarget=(String)e.nextElement();
                putData(nextTarget);
              }
            }
            else{return COMPETEION;}

          //if the target actor is not in the nonLocalReachable Set,
          //it is assumed live.
          } else {
            return FAILURE;
          }
        } else {return ERROR;}
      }
      if (totalTargets==0) {return SUCCESS;}
      else {return MIGRATION;}
    }
    return ERROR;
  }

  //if null is return, there is a mistake here,
  // end the global mark & sweep
  public String getNextHost() {
    String localHost=getLocalLocation();
    for (Enumeration e = tracingList.keys();  e.hasMoreElements(); ) {
      String nextHost=(String)e.nextElement();
      if (nextHost.equals(localHost)) {continue;}
      HashSet data=(HashSet)tracingList.get(nextHost);
      if (data!=null && data.size()>0){return nextHost;}
    }
    GCFail();
    return null;
  }

  public void GCFail() {}

  public void GCSucceed() {}

  public void putData(String targetAddr) {
    String host=getLocation(targetAddr);
    HashSet data=(HashSet)tracingList.get(host);
    if (data==null) {
      data=new HashSet();
      if (data.add(targetAddr)) {totalTargets++;}
      tracingList.put(host,data);
    } else {
      if (data.add(targetAddr)) {totalTargets++;}
    }
  }

  public void removeData(String targetAddr) {
    String host=getLocation(targetAddr);
    HashSet data=(HashSet)tracingList.get(host);
    if (data!=null && data.remove(targetAddr)) {totalTargets--;}
    else {System.err.println("SALSA Global GC Failed!");}
  }

  private String getLocation(String targetAddr) {
    if (targetAddr.charAt(0)=='u') { return (new UAN(targetAddr)).getLocation();}
    else if (targetAddr.charAt(0)=='r'){return (new UAL(targetAddr)).getLocation();}
    return null;
  }

  public static String getLocalLocation() {return ServiceFactory.getTheater().getLocation();}

  private void readObject(java.io.ObjectInputStream in) throws java.io.IOException, java.lang.ClassNotFoundException{
    in.defaultReadObject();
    gc=ServiceFactory.getGC();
  }



}