package gc.serverGC;

/**
 * <p>Title: The 'N'on-'L'ocal reachability version of the 'S'erver-client
 *           'Local' 'P'seudo 'R'oot 'ID'entification garbage collector</p>
 * <p>Description: 1. Serves as a normal local collector
 *                 2. Serves as a "server" to provide local snapshot
 *                    to the "client" of the logically centralized
 *                    mobile global GC agent.
 *                    The "client" (global GC agent) gathers information of all
 *                    local snapshot and merge them.
 *                 3. The local gabage finder uses the Pseudo-Root
 *                    IDentifincation (PRID) algorithm. There are two kinds of
 *                    pseudo roots. The first one is the active pseudo root. The
 *                    second is the outside (the snapshot) pseudo root.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: WWC</p>
 * @author WeiJen Wang
 * @version 1.0
 */


import java.lang.ClassCastException;
import java.util.Hashtable;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Vector;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import wwc.naming.WWCNamingService;
import salsa.language.*;
import salsa.naming.UAL;
import salsa.naming.UAN;
import gc.*;
import gc.message.RemoveInverseRefMsg;


public class NLSLocalPRID implements LocalCollector{

//the local naming service
  private WWCNamingService nameSvc;

  //the actors in the snapshot
  private Hashtable actorArray;

  //the internal data structure for mark and sweep
  private LinkedList BFSMemVector;

  //the temporary storage for the system massage "RemoveInverseRefMsg"
  private Vector removeInvRefTarget;

  //the garbage array identified by the prid algorithm
  private Hashtable garbageArray;

  private int throughPut=0;

  private String lock=null;

  public class invTargetWrap {
    public String target;
    public String argument;
    int debt;
    public invTargetWrap(String target,String arg,int debt) {
      this.target=target;this.argument=arg;this.debt=debt;
    }
  }

  public NLSLocalPRID() {
    try {
      nameSvc = (WWCNamingService) ServiceFactory.getNaming();
    } catch (ClassCastException e) {nameSvc=null;}
    actorArray=new Hashtable();
    BFSMemVector=new LinkedList();
    removeInvRefTarget=new Vector();
    garbageArray=new Hashtable();
    throughPut=0;
  }

  public  boolean occupy(String detectiveName) {return false;}
  public  void free(String detectiveName) {}
  public  Hashtable getNonlocalReachableSet(String detectiveName) {return null;}

  public boolean requestSnapShot(String name) {
    if (!lockGC(name)) return false; //snapshot failed
    formGroup();
    processSystemMessages();
    snapshot();
    return true;
  }

  public Vector requestVirtualMigration(String name) {
    if (!lockGC(name)) return null; //request failed

    Vector virtualMigrationList=new Vector();

    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor=(UniversalActor.State)e.nextElement();
      ActorMemory currentMemory=actor.getActorMemory();

      //if the actor is unblocked, color it
      //if it is wating for an ACK, color it.
      if (currentMemory.isActive() || currentMemory.isLocked() || currentMemory.GCGetForwardList().isPseudoRoot()) {
        VirtualActor va=new VirtualActor(currentMemory.GCGetForwardList().getSelfRef(),
                                         true,
                                         currentMemory.GCGetForwardList().getRealForwardList(),
                                         currentMemory.GCGetinverseList().getRealInverseList());
        virtualMigrationList.addElement(va);
      }
      else {
        VirtualActor va=new VirtualActor(currentMemory.GCGetForwardList().getSelfRef(),
                                         false,
                                         currentMemory.GCGetForwardList().getRealForwardList(),
                                         currentMemory.GCGetinverseList().getRealInverseList());
        virtualMigrationList.addElement(va);
      }
    }
    GCFailed();
    return virtualMigrationList;
  }

  public boolean requestKilling(String name,Vector victimList) {
    if (!lockGC(name)) return false; //request failed
    this.throughPut=0;
    formGroup();
    processSystemMessages();
    snapshot();

    for (int i=0;i<victimList.size();i++) {
        UniversalActor.State garbageActor=(UniversalActor.State) actorArray.get((String)victimList.get(i));
        ActorMemory currentMem;
        if (garbageActor!=null) {
             this.throughPut++;
            currentMem = garbageActor.getActorMemory();
            garbageArray.put(currentMem.GCGetForwardList().getSelfRef(),garbageActor);
            removeAcquaintanceInvRef(currentMem.GCGetForwardList(), true);
        }
    }
    collectGarbage();
    HashSet garbage=new HashSet(victimList);

    //remove extra inverse reference deletion messages
    for (int i=0;i<removeInvRefTarget.size();i++) {
      invTargetWrap data=(invTargetWrap)removeInvRefTarget.get(i);
      if (!garbage.contains(data.target)) {
        //send system messages to the inverse acquaintances
        //to remove the inverse references
        WeakReference invActor;
        if (data.target.charAt(0)=='u') {invActor = new WeakReference(new UAN(data.target), null);}
        else {invActor = new WeakReference(null, new UAL(data.target));}
        invActor.send(new RemoveInverseRefMsg(invActor,data.argument, data.debt));
      }
    }

    freeLock(name);
    GCFailed();
System.out.println("Global GC throughPut="+this.throughPut);
    return true;
  }

  private synchronized boolean lockGC(String name) {
    if (this.lock==null) {this.lock=name;return true;}
    else if (this.lock.equals(name)) {return true;}
    else {return false;}

  }

  private synchronized void freeLock(String name) {
    if (name.equals(this.lock)) { this.lock=null;}
  }










/**********************************************
*   the pure local part
***********************************************/
  public void collect() {
    if (!lockGC("me")) {
      return;//fail to gc;
    }
    System.gc();
    formGroup();
    processSystemMessages();
    snapshot();
    pseudoRootIdentification();
    globalRootIdentification();
    BFSMark();
    getGarbageActorAndAcqInvRef();
    condenseForwardReference();
    collectGarbage();
    GCFinalize();
    freeLock("me");
  }

  public int getThroughPut() {return throughPut;}

  protected void condenseForwardReference() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor = (UniversalActor.State) e.nextElement();
      ActorMemory currentMemory = actor.getActorMemory();
      removeAcquaintanceInvRef(currentMemory.GCGetForwardList(),false);
    }
  }

  protected void processSystemMessages(){
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor=(UniversalActor.State)e.nextElement();
      for (;;) {
        try {
          Message msg = (Message) actor.sysMailbox.remove(0);
          if (msg.getMethodName().equals("receiveAck")) {
            actor.getActorMemory().GCGetForwardList().receiveAck((String)(msg.getArguments())[0]);
          } else if (msg.getMethodName().equals("removeForwardRef")) {
            actor.getActorMemory().GCGetForwardList().removeReference((String)(msg.getArguments())[0]);
          } else if (msg.getMethodName().equals("removeInverseRef")) {
            actor.getActorMemory().GCGetinverseList().removeInverseReference((String)(msg.getArguments())[0],((Integer)(msg.getArguments())[1]).intValue());
          }
        } catch (java.lang.ArrayIndexOutOfBoundsException aie) {break;}
      }
    }
  }

  protected void formGroup() {
   snapshotUANReg(nameSvc.getUANTable());
   snapshotUALReg(nameSvc.getUALTable());
   BFSMemVector.clear();
   removeInvRefTarget.clear();
   garbageArray.clear();
  }

  protected void snapshot() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor=(UniversalActor.State)e.nextElement();
      actor.getActorMemory().startGC();
    }
  }



  protected void pseudoRootIdentification() {

      for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
        UniversalActor.State actor=(UniversalActor.State)e.nextElement();
        ActorMemory currentMemory=actor.getActorMemory();

        //if the actor is unblocked, color it
        if (currentMemory.isActive() || currentMemory.isLocked() ||actor.currentMessage!=null) {
          currentMemory.colorPseudoRoot();
          BFSMemVector.addLast(currentMemory);

        }

        //if it is wating for an ACK, color it.
        else if (currentMemory.GCGetForwardList().isPseudoRoot()) {
          currentMemory.colorPseudoRoot();
          BFSMemVector.addLast(currentMemory);

        }
      }
  }

  protected void globalRootIdentification() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor=(UniversalActor.State)e.nextElement();
      ActorMemory currentMemory=actor.getActorMemory();

      // One step back tracking:
      //   If an actor has an inverse acquaintance outside
      //   the snapshot group, color it.

        Hashtable ilist = currentMemory.GCGetinverseList().getRealInverseList();
        if (ilist==null) {continue;}
        for (Enumeration invRefStrEnu = ilist.keys();
             invRefStrEnu.hasMoreElements(); ) {
          String searchKey = (String) invRefStrEnu.nextElement();
          UniversalActor.State inverseAcquaintance=(UniversalActor.State)actorArray.get(searchKey);
          if (inverseAcquaintance == null) {
            currentMemory.colorPseudoRoot();
            BFSMemVector.addLast(currentMemory);
            break;
          }
        }
    }
  }



  protected void BFSMark() {
    ActorMemory currentMem;
    for (;;) {

      try {currentMem = (ActorMemory) BFSMemVector.removeFirst();}
      catch (NoSuchElementException e) {break;}

      Hashtable flist=currentMem.GCGetForwardList().getRealForwardList();
      for (Enumeration e = flist.keys();  e.hasMoreElements(); ) {
        String nextActorStr=(String)e.nextElement();
        ForwardListEntry fEntry=(ForwardListEntry)flist.get(nextActorStr);

        //if (fEntry.getReferenceCounter()<=0) {continue;}
        Object nextActorObj=actorArray.get(nextActorStr);
        if (nextActorObj!=null) {
          UniversalActor.State nextActor = (UniversalActor.State) nextActorObj;
          ActorMemory nextMem = nextActor.getActorMemory();
          if (nextMem.isLocalMarked() || nextMem.isPseudoRoot()) {
            continue;
          } else {
            nextMem.colorLocalMark();
            BFSMemVector.addLast(nextMem);
          }
        }
      }
    }
    BFSMemVector.clear();
  }

  protected void getGarbageActorAndAcqInvRef() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State garbageActor=(UniversalActor.State) e.nextElement();
      ActorMemory currentMem=garbageActor.getActorMemory();
      if (!currentMem.isLocalMarked() && !currentMem.isPseudoRoot()) {
        garbageArray.put(currentMem.GCGetForwardList().getSelfRef(),garbageActor);
        removeAcquaintanceInvRef(currentMem.GCGetForwardList(),true);
      }
      removeAcquaintanceInvRef(currentMem.GCGetForwardList(),false);
    }
  }


  //try to remove useless forward references and put them in 'removeInvRefTarget'
  private void removeAcquaintanceInvRef(ForwardList fList, boolean killed) {
    //find all unused forward references
    for (Enumeration e = fList.getRealForwardList().keys();  e.hasMoreElements(); ) {
      String key=(String) e.nextElement();
      Hashtable flistTable=fList.getRealForwardList();
      ForwardListEntry fEntry=(ForwardListEntry) flistTable.get(key);

      //If the forward reference is useless, remove it, and store their information in 'removeInvRefTarget'
      //The forward references are useless if
      //  1. the source actor is live but some of its forward references are dead
      //  2. the source actor is garbage.
      //Notice that is killed=true, every forward reference is removed.
      if (fEntry.isDead() || killed) {
         Object res=flistTable.remove(key);
         if (res!=null) {
           //put the inverse reference information in 'removeInvRefTarget'
           invTargetWrap invData=new invTargetWrap(key,fList.getSelfRef(),fEntry.getDebt());
           removeInvRefTarget.addElement(invData);
         }
      }
    }
  }

  protected void collectGarbage() {
    throughPut=0;
    for (Enumeration e = garbageArray.elements();  e.hasMoreElements();) {
      UniversalActor.State gActor=((UniversalActor.State)e.nextElement());
      nameSvc.remove(gActor.getUAN(),gActor.getUAL());
      gActor.GCdie();
      throughPut++;
    }
  }

  protected void GCFinalize() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor = (UniversalActor.State)e.nextElement();
      try {actor.getActorMemory().endGC();}
      catch (Exception exc) {
        System.err.println("GC debug info:"+e);
        continue;
      }
    }
    notifyInvAcq();
    this.removeInvRefTarget.clear();
    this.garbageArray.clear();
//this.showActorState();
    this.actorArray.clear();
    this.BFSMemVector.clear();
    System.gc();
  }

  protected void GCFailed() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor = (UniversalActor.State)e.nextElement();
      try {actor.getActorMemory().endGC();}
      catch (Exception exc) {
        System.err.println("GC debug info:"+e);
        continue;
      }
    }
    this.removeInvRefTarget.clear();
    this.garbageArray.clear();
    this.actorArray.clear();
    this.BFSMemVector.clear();
    System.gc();
  }

  /**
   * firstly, remove all unnecessary msgs whose targets are garbage.
   * Then start sending messages to unregister inverse references
   * This function is used in GCFinalize()
   */
  private void notifyInvAcq() {
    //remove all unnecessary system messages which will be sent to a garbage actor
    for (int i=0;i<removeInvRefTarget.size();i++) {
      invTargetWrap data=(invTargetWrap)removeInvRefTarget.get(i);
      if (garbageArray.get(data.target)==null) {
        //send system messages to the inverse acquaintances
        //to remove the inverse references
        WeakReference invActor;
        if (data.target.charAt(0)=='u') {invActor = new WeakReference(new UAN(data.target), null);}
        else {invActor = new WeakReference(null, new UAL(data.target));}
        invActor.send(new RemoveInverseRefMsg(invActor,data.argument, data.debt));
      }
    }
  }



  //pass the uan registry hashtable that stores actors
  // to find them and put them in the 'actorArray'
  private void snapshotUANReg(Hashtable reg) {
      for (Enumeration e = reg.elements();  e.hasMoreElements(); ) {
        Object actor = e.nextElement();
        if (actor instanceof UniversalActor.State) {
          if (actor instanceof salsa.resources.EnvironmentalServiceState) {continue;}
          if (actor instanceof salsa.resources.ActorServiceState) {continue;}
          actorArray.put(((UniversalActor.State)actor).getUAN().toString(),actor);
          //( (UniversalActor.State) actor).getActorMemory().startGC();
        }
      }
  }

  private void snapshotUALReg(Hashtable reg) {
      for (Enumeration e = reg.elements();  e.hasMoreElements(); ) {
        Object actor = e.nextElement();
        if (actor instanceof UniversalActor.State) {
          if (actor instanceof salsa.resources.EnvironmentalServiceState) {continue;}
          if (actor instanceof salsa.resources.ActorServiceState) {continue;}
          actorArray.put(((UniversalActor.State)actor).getUAL().toString(),actor);
          //( (UniversalActor.State) actor).getActorMemory().startGC();
        }
      }
  }


  //debugging tool:
  public void showActorState() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor=(UniversalActor.State)e.nextElement();
      System.out.println(actor.getActorMemory().toString());
    }
  }
}
