package salsa.language;

import gc.*;

/**
 * Runtime class.
 *
 * Initially just keep a counter for unprocessed actor messages.
 * When there are no unprocessed actor messages and all actors are
 * local and are waiting for messages:  we can exit the Java
 * virtual machine.
 */

public class RunTime {
	// Here we have a lower priority thread that checks for
	// conditions to terminate the JVM:
	// (1) All actors are local.  (no universal actors)
	// (2) All actor mailboxes are empty.
	// (3) No actor is currently processing a message.

	static Thread exitThread;
	static int messagesInMailboxes = 0;
	static int outgoingObjects = 0;
	static public int universalActors = 0;
        //static int actors=0;
	static boolean started = false;
	static boolean initialized = false;
        static boolean nogc=false;
        static boolean nodie=false;
        static boolean gcverbose=false;
        //static gc.localGC.LocalPRID gc=new gc.localGC.LocalPRID();
        static LocalCollector gc;

	private static void init() {
          try{
            if (System.getProperty( "nogc" )!=null){nogc=true;}
            if (System.getProperty( "gcverbose" )!=null){gcverbose=true;}
          } catch (Exception e) {nogc=false;}

          if (!nogc) {gc=ServiceFactory.getGC();}
          if (System.getProperty("nodie") == null) {nodie=false;} else {nodie=true;}
		Runnable checkExit = new Runnable() {
			public void run() {
                          long currentMem=Long.MAX_VALUE;
                          int i=0;
					while ( live() || !isDead()) {
						try {
                                                        Thread.sleep(100);
                                                        if (!nogc) {
                                                          long data=Runtime.getRuntime().totalMemory();
                                                          i++;
                                                          if (data>currentMem || i>20) {
                                                            i=0;

                                                            gc.collect();
                                                            if (gcverbose) {
                                                              System.out.println("GC:garbage throughput=" +
                                                                                 gc.getThroughPut() +
                                                                                 ",max=" + Runtime.getRuntime().maxMemory() +
                                                                                 ",total=" +
                                                                                 Runtime.getRuntime().totalMemory() +
                                                                                 ",free=" + Runtime.getRuntime().freeMemory()
                                                                                 );
                                                            }
                                                          }
                                                          currentMem=Runtime.getRuntime().totalMemory();
                                                        }
						} catch (Exception ie) {}
					}
//					System.err.println("quitting");
					System.err.flush();
					System.out.flush();

					System.runFinalization();
					System.exit(0);
			}
		};

		exitThread = new Thread(checkExit);
		exitThread.setDaemon(true);
		//exitThread.setPriority(Thread.MIN_PRIORITY);

	}

        //public static boolean passedRef=false;
        public static boolean sentOutgoingObj=false;

        //public static synchronized void setPassRef() {
        //  passedRef=true;
        //}

        public static synchronized boolean live() {
          return nodie || sentOutgoingObj;
        }

        //public synchronized static void createActor() {actors++;}

        //public synchronized static void deleteActor() {actors--;}

	public synchronized static void receivedMessage() {

		messagesInMailboxes++;
                if (!initialized) {
                  initialized = true;
                        init();
                }

		if (!started) {
			started = true;
			exitThread.setName( "RunTime Exit Thead" );
			exitThread.start();
		}
	}

        public synchronized static boolean isDead() {
          return messagesInMailboxes==0  && universalActors==0;
        }

	public synchronized static void finishedProcessingMessage() {
		messagesInMailboxes--;
	}

	// public for access to wwc.messaging package.
	public synchronized static void receivedUniversalActor() {
		universalActors++;
	}

	public synchronized static void removedUniversalActor() {
		universalActors--;
	}

	public  static void sendingOutgoingObject() {
          sentOutgoingObj=true;
		//outgoingObjects++;
	}

	public static void sentOutgoingObject() {
		//outgoingObjects--;
	}
}
